unit Game;

interface

uses
  Classes, Forms, Fields, RasterField, Player, Globals;

type TGame = class
               HistoryData: TMemoryStream;
               HistoryPos, HistoryEntries: integer;
               HistoryEntrySize: integer;
             public
               Raster: TRaster;
               Field: TField;
               CurrentPlayer: integer;
               Player: array [0..1] of TPlayer;
               Paused: boolean;
               procedure Restart;
               function CommitMove(Move:TMove): boolean;
               procedure NextPlayer;
               // History-Verwaltung
               procedure Push;
               function Pop: boolean;
               function Go(Destination: integer): boolean;
               function Xdo(UndoRedo: integer): boolean;
             end;

var TheGame: TGame;

implementation

uses
  Main, Dialogs;

procedure TGame.Restart;
var Name   : string;
    Human  : boolean;
    AILevel: integer;
begin
  if Assigned(Field) then Field.Free;
  Field:=TField.Create(Raster);

  if Assigned(Player[0]) then begin
    Name   := Player[0].Name;
    Human  := Player[0].IsHuman;
    AILevel:= Player[0].AILevel;
    Player[0].Free;
  end else begin
    Name   := 'Spieler 1';
    Human  := true;
    AILevel:= 0;
  end;
  TheGame.Player[0] := TPlayer.Create(0, Raster, Name, Human, AILevel);

  if Assigned(Player[1]) then begin
    Name   := Player[1].Name;
    Human  := Player[1].IsHuman;
    AILevel:= Player[1].AILevel;
    Player[1].Free;
  end else begin
    Name   := 'Spieler 2';
    Human  := true;
    AILevel:= 0;
  end;
  TheGame.Player[1] := TPlayer.Create(1, Raster, Name, Human, AILevel);

  Field.DeterminePossibleOwners;
  HistoryPos := -1;
  HistoryEntries := 0;
  HistoryEntrySize := sizeof(THistoryHeader) + 3 * Field.DataSize;
  if not Assigned(HistoryData) then
    HistoryData := TMemoryStream.Create;
  Paused := false;

  with MainForm do begin
    BPause.Visible := not(Player[0].IsHuman OR Player[1].IsHuman);
    BContinue.Visible := false;
  end;

  CurrentPlayer := 1; // <Kinkel> ich wei... :-/ <== <KeyJ> *LOOOOOOOOL*
  Push;
  NextPlayer;
end;

procedure TGame.Push;
var NeededSize: integer;
    Header: THistoryHeader;
begin
  inc(HistoryPos);
  HistoryEntries := HistoryPos + 1;
  NeededSize := HistoryEntries * HistoryEntrySize;
  if HistoryData.Size < NeededSize then
    HistoryData.SetSize(NeededSize + 8 * HistoryEntrySize);
  Header.CurrentPlayer := CurrentPlayer;
  Header.RemainingMoves[0] := Player[0].RemainingMoves;
  Header.RemainingMoves[1] := Player[1].RemainingMoves;
  HistoryData.Position := HistoryPos * HistoryEntrySize;
  HistoryData.Write(Header, sizeof(Header));
  with Field do
    HistoryData.Write(DataPtr^, DataSize);
  with Player[0].RemainingField do
    HistoryData.Write(DataPtr^, DataSize);
  with Player[1].RemainingField do
    HistoryData.Write(DataPtr^, DataSize);
end;

function TGame.Pop: boolean;
begin
  Result := Go(HistoryPos);
  if Result then dec(HistoryPos);
end;

function TGame.Go(Destination: integer): boolean;
var Header: THistoryHeader;
begin
  if (Destination < 0) OR (Destination >= HistoryEntries) then begin
    Result := false;
    exit;
  end else
    Result := true;
  HistoryPos := Destination;
  HistoryData.Position := Destination * HistoryEntrySize;
  HistoryData.Read(Header, sizeof(Header));
  CurrentPlayer := Header.CurrentPlayer;
  Player[0].RemainingMoves := Header.RemainingMoves[0];
  Player[1].RemainingMoves := Header.RemainingMoves[1];
  with Field do
    HistoryData.Read(DataPtr^, DataSize);
  with Player[0].RemainingField do
    HistoryData.Read(DataPtr^, DataSize);
  with Player[1].RemainingField do
    HistoryData.Read(DataPtr^, DataSize);
end;

function TGame.Xdo(UndoRedo: integer): boolean;
begin
  if Player[0].IsHuman XOR Player[1].IsHuman then
    UndoRedo := UndoRedo * 2;
  Result := Go(HistoryPos + UndoRedo);
  if not Result then exit;
  NextPlayer;
end;

function TGame.CommitMove(Move: TMove): boolean;
begin
  Result:=Field.MoveCheck(Move,[mcCommit]);
  if Result then begin
    Field.DeterminePossibleOwners(Move);
    Player[0].DetermineBlockedMoves(Move, [dbmCommit]);
    Player[1].DetermineBlockedMoves(Move, [dbmCommit]);
    Push;
    NextPlayer;
  end;
end;

procedure TGame.NextPlayer;
var Move: TMove;
begin
  if Paused then begin
    MainForm.NotifyUpdate;
    exit;
  end else begin
    CurrentPlayer := CurrentPlayer XOR 1;
    MainForm.NotifyUpdate;
  end;

  if not Player[CurrentPlayer].CanMove then begin
    MessageDlg(Player[CurrentPlayer XOR 1].Name + ' hat gewonnen.',
               mtInformation, [mbOK], 0);
    MainForm.MHint.Enabled := false;
    exit;
  end;

  MainForm.MHint.Enabled := Player[CurrentPlayer].IsHuman;

  if not Player[CurrentPlayer].IsHuman then begin
    Move := Player[CurrentPlayer].DetermineMove;
    CommitMove(Move);
  end;
end;

end.
