unit Globals;

interface

const BuildID='2004-01-29-00-10';
      Creator='KeyJ/Kinkel/Rotfuchs';
      //----------------------------//
      OWNER_OOF       = -2;
      OWNER_EMPTY     = -1;
      OWNER_DEAD      =  2;
      OWNER_UNDECIDED = OWNER_EMPTY;
      //----------------------------//
      DIR_H   =  1;  { - }
      DIR_V   =  2;  { | }
      DIR_D   =  4;  { / }
      DIR_A   =  8;  { \ }
      ALL_DIR = 15;  { * } // ;-)
      NO_DIR  =  0;
      //----------------------------//
      LINES_SURROUNDING = 12;
      LINES_EXTENDING   =  3;
      //----------------------------//
      xdoUndo = -1;
      xdoRedo = +1;
      //----------------------------//
      GAMETREE_BRANCHFACTOR = 5;
      GAMETREE_DEPTH        = 5;
      INFIMUM_POS           =  32767;
      INFIMUM_NEG           = -32768;

type TMove  = record
                X,Y    : integer;  // Ein Zug ist gekennzeichnet durch die
                Dir    : integer;  // Position (X,Y), die Richtung (Dir) und
                Player : integer;  // den Spieler (Player), der den Zug ausfhrt
              end;
     PNode  = ^TNode;
     TNode  = record
                Owner          : integer;
                PossibleOwners : integer;
                Dirs           : integer;
                PossibleDirs   : integer;
              end;
     TDelta = record
                DeltaX : integer;
                DeltaY : integer;
                Dir    : integer;
              end;
     TDeltaList =
              record
                Count : integer;
                Items : array [1..36] of TDelta;
              end;
     TDeltaTable =
              array [DIR_H..DIR_A] of TDeltaList;
     THistoryHeader =
              record
                CurrentPlayer  : integer;
                RemainingMoves : array [0..1] of integer;
              end;
     TRatedMove =
              record
                Rating : integer;
                Move   : TMove;
              end;
     TRatedMoves =
              array [1..GAMETREE_BRANCHFACTOR] of TRatedMove;

type TDBMFlag        = (dbmCommit);
     TDBMFlags       = set of TDBMFlag;
     TMoveCheckFlag  = (mcCommit, mcSoftTest);
     TMoveCheckFlags = set of TMoveCheckFlag;
     TUIState        = (uisStartup, uisDesign, uisGame);

function DoMovesCross(Move1, Move2: TMove): boolean;
  // prft, ob sich 2 Linien berhren / schneiden
function AddDelta(Move: TMove; D: TDelta): TMove;
  // "verschiebt" einen Move in die Richtung des LookupTable-Eintrags
function NextMove(var Move: TMove): boolean;
  // liefert den "sukzessive" nchsten Move zurck, falls er noch innerhalb
  // der Feldgrenzen ist bzw. berhauot mglich ist (bezieht sich dabei
  // auf das RemainingField des Spielers, der Move ausfhren soll
function Minimize(Opponent: integer; Move: TMove;
                  Depth: integer): integer; overload;
  // berechnet das Minimum des durch diesen Zug erreichten Wertes der Heuristik
function Maximize(Opponent: integer; Move: TMove;
                  Depth: integer): integer; overload;
  // berechnet das Maximum des durch diesen Zug erreichten Wertes der Heuristik
function Minimize(Opponent: integer; Move: TMove;
                  Depth, Alpha, Beta: integer): integer; overload;
  // analog mit Parametern fr Alpha-Beta-Heuristik
function Maximize(Opponent: integer; Move: TMove;
                  Depth, Alpha, Beta: integer): integer; overload;
  // analog mit Parametern fr Alpha-Beta-Heuristik

var AbortMove: boolean;  // wenn true, wird der KI-Zug sofort abgebrochen


implementation

uses
  Game, Tables, Progress, Math, Forms;

function DoMovesCross(Move1, Move2: TMove): boolean;
var TmpMove : TMove;
    i : integer;
    PNd1, PNd2 : array [1..LINES_EXTENDING] of PNode;
begin
  Result := false;
  if Move1.Dir > 0 then with TheGame.Field do begin
    for i := 1 to 3 do begin
      TmpMove := AddDelta(Move1, DeltaTable[Move1.Dir].Items[i]);
      PNd1[i] := FieldPtr[TmpMove.X,TmpMove.Y];
      TmpMove := AddDelta(Move2, DeltaTable[Move2.Dir].Items[i]);
      PNd2[i] := FieldPtr[TmpMove.X,TmpMove.Y];
    end;
    Result := (PNd1[1]=PNd2[1]) or (PNd1[1]=PNd2[2]) or (PNd1[1]=PNd2[3]) or
              (PNd1[2]=PNd2[1]) or (PNd1[2]=PNd2[2]) or (PNd1[2]=PNd2[3]) or
              (PNd1[3]=PNd2[1]) or (PNd1[3]=PNd2[2]) or (PNd1[3]=PNd2[3]);
    // nur bei entgegengesetzten Diagonalen noch ein Schnitt mglich, und zwar genau
    // dann, wenn die Mittelpunkte der Linien auf benachbarten Rasterpunkten liegen
    if (Move1.Dir or Move2.Dir) = (DIR_D or DIR_A) then
      Result := Result or ((abs(Move1.X-Move2.X)=1) xor (abs(Move1.Y-Move2.Y)=1));
  end;
end;

function AddDelta(Move: TMove; D: TDelta): TMove;
begin
  Result.X      := Move.X + D.DeltaX;
  Result.Y      := Move.Y + D.DeltaY;
  Result.Dir    := D.Dir;
  Result.Player := Move.Player;
end;

function NextMove(var Move: TMove): boolean;
begin
  Result := true;
  with TheGame.Player[Move.Player].RemainingField do begin
    if Move.Dir >= DIR_A then begin
      Move.Dir := DIR_H;
      if Move.X >= Width - 1 then begin
        Move.X := 0;
        if Move.Y >= Height - 1 then begin
          Result := false;
          exit;
        end else inc(Move.Y);
      end else inc(Move.X);
    end else Move.Dir := Move.Dir shl 1;

    // isser berhaupt gltig ?
    if ((FieldPtr[Move.X, Move.Y]^.PossibleDirs AND Move.Dir)  = 0) or
       ((FieldPtr[Move.X, Move.Y]^.Dirs         AND Move.Dir) <> 0) then
      Result := NextMove(Move);
  end;
end;

function Minimize(Opponent: integer; Move: TMove; Depth: integer): integer;
var Count, i: integer;
    RatedMoves: TRatedMoves;
begin
  Result := INFIMUM_POS;
  with TheGame do begin
    // Spielfeld sichern
    Push;
    if Field.MoveCheck(Move,[mcCommit]) then begin
      // ntige Aktualisierungen am Spielfeld vornehmen
      Field.DeterminePossibleOwners(Move);
      Player[0].DetermineBlockedMoves(Move, [dbmCommit]);
      Player[1].DetermineBlockedMoves(Move, [dbmCommit]);
      // Tiefenschnitt
      if Depth >= GAMETREE_DEPTH then begin
        Result := Player[Opponent].RemainingMoves;
        Pop;
        ProgressForm.ProgressBar.StepIt;
        exit;
      end;
      // heuristische Bewertung fr Count beste Moves des nchsten Spielers
      // (...xor 1) ermitteln --> die Begrenzung der Breite des Spielbaumes
      // ist durch die Lnge des Arrays RatedMoves gegeben
      Player[Move.Player xor 1].DetermineBestRatedMoves(Count, RatedMoves);
      // Falls Spieler keine Zge mehr ausfhren kann, abbrechen, sonst
      // gibt's ne Endlosschleife
      if Count = 0 then begin
        Result := 0;
        Pop;
        ProgressForm.ProgressBar.StepBy(Trunc(
            IntPower(GAMETREE_BRANCHFACTOR, GAMETREE_DEPTH - Depth)));
        exit;
      end;
      // Count Ergebnisse minimieren
      for i := 1 to Count do begin
        Result := Min(Result,
                      Maximize(Opponent, RatedMoves[i].Move, Depth + 1));
        if AbortMove then break;
      end;
    end;
    // Spielfeld wiederherstellen
    Pop;
    // ProgressBar aktualisieren
    if Count < GAMETREE_BRANCHFACTOR then ProgressForm.ProgressBar.StepBy(
      Trunc(IntPower(GAMETREE_BRANCHFACTOR, GAMETREE_DEPTH - Depth)) - Count *
      Trunc(IntPower(GAMETREE_BRANCHFACTOR, GAMETREE_DEPTH - Depth - 1)));
  end;
  Application.ProcessMessages;
end;

function Maximize(Opponent: integer; Move: TMove; Depth: integer): integer;
var Count, i: integer;
    RatedMoves: TRatedMoves;
begin
  Result := INFIMUM_NEG;
  with TheGame do begin
    // Spielfeld sichern
    Push;
    if Field.MoveCheck(Move,[mcCommit]) then begin
      // ntige Aktualisierungen am Spielfeld vornehmen
      Field.DeterminePossibleOwners(Move);
      Player[0].DetermineBlockedMoves(Move, [dbmCommit]);
      Player[1].DetermineBlockedMoves(Move, [dbmCommit]);
      // Tiefenschnitt
      if Depth >= GAMETREE_DEPTH then begin
        Result := Player[Opponent].RemainingMoves;
        Pop;
        ProgressForm.ProgressBar.StepIt;
        exit;
      end;
      // heuristische Bewertung fr Count beste Moves des nchsten Spielers
      // (...xor 1) ermitteln --> die Begrenzung der Breite des Spielbaumes
      // ist durch die Lnge des Arrays RatedMoves gegeben
      Player[Move.Player xor 1].DetermineBestRatedMoves(Count, RatedMoves);
      // Falls Spieler keine Zge mehr ausfhren kann, abbrechen, sonst
      // gibt's ne Endlosschleife
      if Count = 0 then begin
        Result := 0;
        Pop;
        ProgressForm.ProgressBar.StepBy(Trunc(
            IntPower(GAMETREE_BRANCHFACTOR, GAMETREE_DEPTH - Depth)));
        exit;
      end;
      // Count Ergebnisse maximieren
      for i := 1 to Count do begin
        Result := Max(Result,
                      Minimize(Opponent, RatedMoves[i].Move, Depth + 1));
        if AbortMove then break;
      end;
    end;
    // Spielfeld wiederherstellen
    Pop;
    // ProgressBar aktualisieren
    if Count < GAMETREE_BRANCHFACTOR then ProgressForm.ProgressBar.StepBy(
      Trunc(IntPower(GAMETREE_BRANCHFACTOR, GAMETREE_DEPTH - Depth)) - Count *
      Trunc(IntPower(GAMETREE_BRANCHFACTOR, GAMETREE_DEPTH - Depth - 1)));
  end;
  Application.ProcessMessages;
end;

function Minimize(Opponent: integer; Move: TMove;
                  Depth, Alpha, Beta: integer): integer;
var Count, i: integer;
    RatedMoves: TRatedMoves;
begin
  with TheGame do begin
    // Spielfeld sichern
    Push;
    if Field.MoveCheck(Move,[mcCommit]) then begin
      // ntige Aktualisierungen am Spielfeld vornehmen
      Field.DeterminePossibleOwners(Move);
      Player[0].DetermineBlockedMoves(Move, [dbmCommit]);
      Player[1].DetermineBlockedMoves(Move, [dbmCommit]);
      // Tiefenschnitt
      if Depth >= GAMETREE_DEPTH then begin
        Result := Player[Opponent].RemainingMoves;
        Pop;
        ProgressForm.ProgressBar.StepIt;
        exit;
      end;
      // heuristische Bewertung fr Count beste Moves des nchsten Spielers
      // (...xor 1) ermitteln --> die Begrenzung der Breite des Spielbaumes
      // ist durch die Lnge des Arrays RatedMoves gegeben
      Player[Move.Player xor 1].DetermineBestRatedMoves(Count, RatedMoves);
      // Falls Spieler keine Zge mehr ausfhren kann, abbrechen, sonst
      // gibt's ne Endlosschleife
      if Count = 0 then begin
        Result := 0;
        Pop;
        ProgressForm.ProgressBar.StepBy(Trunc(
            IntPower(GAMETREE_BRANCHFACTOR, GAMETREE_DEPTH - Depth)));
        exit;
      end;
      // Count Ergebnisse minimieren unter Bercksichtigung der Grenzen
      for i := 1 to Count do begin
        Beta := Min(Beta, Maximize(Opponent, RatedMoves[i].Move, Depth + 1,
                                   Alpha, Beta));
        if Beta < Alpha then begin
          // Alpha-Schnitt
          Result := Alpha;
          Pop;
          ProgressForm.ProgressBar.StepBy(
            Trunc(IntPower(GAMETREE_BRANCHFACTOR, GAMETREE_DEPTH - Depth)) - i *
            Trunc(IntPower(GAMETREE_BRANCHFACTOR, GAMETREE_DEPTH - Depth - 1)));
          exit;
        end;
        if AbortMove then break;
      end;
      Result := Beta;
    end else Result := INFIMUM_POS;
    // Spielfeld wiederherstellen
    Pop;
    // ProgressBar aktualisieren
    if Count < GAMETREE_BRANCHFACTOR then ProgressForm.ProgressBar.StepBy(
      Trunc(IntPower(GAMETREE_BRANCHFACTOR, GAMETREE_DEPTH - Depth)) - Count *
      Trunc(IntPower(GAMETREE_BRANCHFACTOR, GAMETREE_DEPTH - Depth - 1)));
  end;
  Application.ProcessMessages;
end;

function Maximize(Opponent: integer; Move: TMove;
                  Depth, Alpha, Beta: integer): integer;
var Count, i: integer;
    RatedMoves: TRatedMoves;
begin
  with TheGame do begin
    // Spielfeld sichern
    Push;
    if Field.MoveCheck(Move,[mcCommit]) then begin
      // ntige Aktualisierungen am Spielfeld vornehmen
      Field.DeterminePossibleOwners(Move);
      Player[0].DetermineBlockedMoves(Move, [dbmCommit]);
      Player[1].DetermineBlockedMoves(Move, [dbmCommit]);
      // Tiefenschnitt
      if Depth >= GAMETREE_DEPTH then begin
        Result := Player[Opponent].RemainingMoves;
        Pop;
        ProgressForm.ProgressBar.StepIt;
        exit;
      end;
      // heuristische Bewertung fr Count beste Moves des nchsten Spielers
      // (...xor 1) ermitteln --> die Begrenzung der Breite des Spielbaumes
      // ist durch die Lnge des Arrays RatedMoves gegeben
      Player[Move.Player xor 1].DetermineBestRatedMoves(Count, RatedMoves);
      // Falls Spieler keine Zge mehr ausfhren kann, abbrechen, sonst
      // gibt's ne Endlosschleife
      if Count = 0 then begin
        Result := 0;
        Pop;
        ProgressForm.ProgressBar.StepBy(Trunc(
            IntPower(GAMETREE_BRANCHFACTOR, GAMETREE_DEPTH - Depth)));
        exit;
      end;
      // Count Ergebnisse maximieren unter Bercksichtigung der Grenzen
      for i := 1 to Count do begin
        Alpha := Max(Alpha, Minimize(Opponent, RatedMoves[i].Move, Depth + 1,
                                     Alpha, Beta));
        if Alpha > Beta then begin
          // Beta-Schnitt
          Result := Beta;
          Pop;
          ProgressForm.ProgressBar.StepBy(
            Trunc(IntPower(GAMETREE_BRANCHFACTOR, GAMETREE_DEPTH - Depth)) - i *
            Trunc(IntPower(GAMETREE_BRANCHFACTOR, GAMETREE_DEPTH - Depth - 1)));
          exit;
        end;
        if AbortMove then break;
      end;
      Result := Alpha;
    end else Result := INFIMUM_NEG;
    // Spielfeld wiederherstellen
    Pop;
    // ProgressBar aktualisieren
    if Count < GAMETREE_BRANCHFACTOR then ProgressForm.ProgressBar.StepBy(
      Trunc(IntPower(GAMETREE_BRANCHFACTOR, GAMETREE_DEPTH - Depth)) - Count *
      Trunc(IntPower(GAMETREE_BRANCHFACTOR, GAMETREE_DEPTH - Depth - 1)));
  end;
  Application.ProcessMessages;
end;

end.