unit RasterField;
interface
uses SysUtils, Classes;

const FieldSaveDir = 'Fields/';
      FieldSaveExt = '.KI22';

type TRaster=class
             private
               SizeX, SizeY: integer;   // Gre
               Data: array of boolean;  // Datenarray
               function GetRaster(X, Y: integer): boolean;  // read Raster
               procedure SetRaster(X, Y: integer; Value: boolean);  // write R.
             public
               // Gre
               property Width: integer read SizeX;
               property Height: integer read SizeY;
               // Zugriff aufs Rasterfeld
               property Raster[X, Y: integer]: boolean
                 read GetRaster write SetRaster; default;
               // Konstruktor/Destruktor
               constructor Create(_Width, _Height: integer); overload;
               destructor Destroy; override;
               // Laden/Speichern von Feldern
               constructor Create(const FieldName: string); overload;
               function Save(const FieldName: string): boolean;
             end;

procedure TRaster_ListSavedFields(List: TStrings);

implementation

constructor TRaster.Create(_Width, _Height: integer);
var Y: integer;
begin
  inherited Create;
  SizeX := _Width;
  SizeY := _Height;
  SetLength(Data, SizeX * SizeY);
  // Feld fllen
  FillChar(Data[SizeX], SizeX * (SizeY - 2) * sizeof(boolean), 1);
  // obere und untere Zeile leeren
  FillChar(Data[0], SizeX * sizeof(boolean), 0);
  FillChar(Data[(SizeY - 1) * SizeX], SizeX * sizeof(boolean), 0);
  // linke und rechte Spalte leeren
  for Y := 1 to (SizeY - 2) do begin
    Data[Y * SizeX] := false;
    Data[(Y + 1) * SizeX - 1] := false;
  end;
end;

destructor TRaster.Destroy;
begin
  Finalize(Data);
  inherited Destroy;
end;

function TRaster.GetRaster(X, Y: integer): boolean;
begin
  if (X >= 0) AND (Y >= 0) AND (X < SizeX) AND (Y < SizeY)
    then Result := Data[Y * SizeX + X]
    else Result := false;
end;

procedure TRaster.SetRaster(X, Y: integer; Value: boolean);
begin
  if (X >= 0) AND (Y >= 0) AND (X < SizeX) AND (Y < SizeY)
    then Data[Y * SizeX + X] := Value;
end;

constructor TRaster.Create(const FieldName: string);
var List: TStringList;
    X, Y: integer;
    Len : integer;
    S   : string;
begin
  try
    List := TStringList.Create();
    List.LoadFromFile(FieldSaveDir+FieldName+FieldSaveExt);
    if (List.Count < 1) OR (length(List[0]) < 1) then
      raise Exception.Create('Ungltige Spielfelddatei');
  except
    SizeX := 1;
    SizeY := 1;
    SetLength(Data, 1);
    Data[0] := false;
    exit;
  end;
  SizeX := length(List[0]);
  SizeY := List.Count;
  SetLength(Data, SizeX * SizeY);
  FillChar(Data[0], SizeX * SizeY * sizeof(boolean), 0);
  for Y:= 0 to (SizeY-1) do begin
    S := List[Y];
    Len := length(S);
    for X := 0 to (Len-1) do
      Raster[X, Y]:= (S[X + 1] = '#');
  end;
end;

function TRaster.Save(const FieldName: string): boolean;
var T: System.Text;
    X, Y: integer;
begin
  {$I-}
  MkDir(FieldSaveDir); IOResult;
  Assign(T, FieldSaveDir+FieldName+FieldSaveExt);
  Rewrite(T);
  {$I+}
  if IOResult <> 0 then begin
    Result := false;
    exit;
  end;
  for Y := 0 to (SizeY-1) do begin
    for X := 0 to (SizeX-1) do
      if Raster[X, Y] then write(T,'#') else write(T,'.');
    writeln(T);
  end;
  CloseFile(T);
  Result := true;
end;

procedure TRaster_ListSavedFields(List: TStrings);
var SR: TSearchRec;
    Cont: integer;
begin
  List.Clear;
  Cont := FindFirst(FieldSaveDir+'*'+FieldSaveExt, $27, SR);
  while Cont = 0 do begin
    List.Add(copy(SR.Name, 1, length(SR.Name) - length(FieldSaveExt)));
    Cont := FindNext(SR);
  end;
  FindClose(SR);
end;

end.
